/*
    Qalculate    

    Copyright (C) 2003  Niklas Knutsson (nq@altern.org)

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.
*/


#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <gtk/gtk.h>
#include <gdk-pixbuf/gdk-pixbuf.h>
#include <unistd.h>
#include <dirent.h>
#include <sys/stat.h>

#include "data/icon.xpm"

#include <libqalculate/qalculate.h>
#include <support.h>

GdkPixbuf *icon_pixbuf;

bool fetch_exchange_rates_at_startup, first_time, first_qalculate_run, canfetch;
string wget_args;
vector<Unit*> currencies;
GtkWidget *currencies_menu_to, *currencies_menu_from;
GtkListStore *currencies_store;
string initial_from_currency, initial_to_currency;

GtkWidget *currencies_entry_to_val;
GtkWidget *currencies_entry_from_val;
GtkWidget *currencies_dialog;

EvaluationOptions evalops;
PrintOptions printops;

gint string_sort_func(GtkTreeModel *model, GtkTreeIter *a, GtkTreeIter *b, gpointer user_data) {
	gint cid = GPOINTER_TO_INT(user_data);
	gchar *gstr1, *gstr2;
	gint retval;
	gtk_tree_model_get(model, a, cid, &gstr1, -1);
	gtk_tree_model_get(model, b, cid, &gstr2, -1);
	retval = g_ascii_strcasecmp(gstr1, gstr2);
	g_free(gstr1);
	g_free(gstr2);
	return retval;
}

void update_currency_menus() {
	GtkTreeIter iter;
	gtk_list_store_clear(currencies_store);
	bool active_from_set = false, active_to_set = false;
	for(size_t i = 0; i < CALCULATOR->units.size(); i++) {
		if(CALCULATOR->units[i]->isActive() && CALCULATOR->units[i]->isCurrency()) {
			gtk_list_store_append(currencies_store, &iter);
			if(!active_from_set && !initial_to_currency.empty() && initial_from_currency == CALCULATOR->units[i]->referenceName()) {
				gtk_combo_box_set_active_iter(GTK_COMBO_BOX(currencies_menu_from), &iter);
				active_from_set = true;
			}
			if(!active_to_set && !initial_to_currency.empty() && initial_to_currency == CALCULATOR->units[i]->referenceName()) {
				gtk_combo_box_set_active_iter(GTK_COMBO_BOX(currencies_menu_to), &iter);
				active_to_set = true;
			}
			string ref = "";
			if(CALCULATOR->units[i]->referenceName().length() == 3) {
				ref = "(";
				ref += CALCULATOR->units[i]->referenceName();
				ref += ")";
			}
			gtk_list_store_set(currencies_store, &iter, 0, CALCULATOR->units[i]->title().c_str(), 1, ref.c_str(), 2, (gpointer) CALCULATOR->units[i], -1);
		}
	}
	
	if(!active_from_set) gtk_combo_box_set_active(GTK_COMBO_BOX(currencies_menu_from), 0);
	if(!active_to_set) gtk_combo_box_set_active(GTK_COMBO_BOX(currencies_menu_to), 0);
}

void load_preferences();
void save_preferences();

#ifdef __cplusplus
extern "C" {
#endif

void convert_currencies(int toFrom = -1);
void fetch_exchange_rates(int timeout);
void on_currencies_button_close_clicked(GtkButton *button, gpointer user_data);
void on_currencies_button_fetch_exchange_rates_clicked(GtkButton *w, gpointer user_data);
void on_currencies_menu_to_changed(GtkComboBox *w, gpointer user_data);
void on_currencies_menu_from_changed(GtkComboBox *w, gpointer user_data);
void on_currencies_entry_from_val_changed(GtkEditable *entry, gpointer user_data);
void on_currencies_entry_to_val_changed(GtkEditable *entry, gpointer user_data);
void close_app();

void fetch_exchange_rates(int timeout) {
#if GTK_MINOR == 6 && GTK_MICRO < 2
	GtkWidget *dialog = gtk_message_dialog_new(GTK_WINDOW(currencies_dialog), GTK_DIALOG_DESTROY_WITH_PARENT, GTK_MESSAGE_INFO, GTK_BUTTONS_CLOSE, _("Fetching exchange rates."));
#else	
	GtkWidget *dialog = gtk_message_dialog_new(GTK_WINDOW(currencies_dialog), GTK_DIALOG_DESTROY_WITH_PARENT, GTK_MESSAGE_INFO, GTK_BUTTONS_NONE, _("Fetching exchange rates."));
#endif	
	gtk_widget_show_now(dialog);
	while(gtk_events_pending()) gtk_main_iteration();
	CALCULATOR->fetchExchangeRates(timeout, wget_args);
	gtk_widget_destroy(dialog);
}
void on_currencies_button_fetch_exchange_rates_clicked(GtkButton*, gpointer) {
	Unit *uFrom = NULL;
	Unit *uTo = NULL;
	GtkTreeIter iter;
	if(gtk_combo_box_get_active_iter(GTK_COMBO_BOX(currencies_menu_from), &iter)) {
		gtk_tree_model_get(GTK_TREE_MODEL(currencies_store), &iter, 2, &uFrom, -1);
	}	
	if(gtk_combo_box_get_active_iter(GTK_COMBO_BOX(currencies_menu_to), &iter)) {
		gtk_tree_model_get(GTK_TREE_MODEL(currencies_store), &iter, 2, &uTo, -1);
	}
	if(uFrom) initial_from_currency = uFrom->referenceName();
	if(uTo) initial_to_currency = uTo->referenceName();
	fetch_exchange_rates(15);
	CALCULATOR->loadExchangeRates();
	update_currency_menus();
	convert_currencies();
}
void on_currencies_button_close_clicked(GtkButton*, gpointer) {
	close_app();
}

void close_app() {
	save_preferences();
	gtk_main_quit();
}

void on_currencies_menu_to_changed(GtkComboBox*, gpointer) {
	convert_currencies();
}
void on_currencies_menu_from_changed(GtkComboBox*, gpointer) {
	convert_currencies();
}
void on_currencies_entry_from_val_changed(GtkEditable*, gpointer) {
	convert_currencies(0);
}
void on_currencies_entry_to_val_changed(GtkEditable*, gpointer) {
	convert_currencies(1);
}

void convert_currencies(int toFrom) {

	GtkTreeIter iter;
	Unit *uFrom = NULL;
	Unit *uTo = NULL;
	if(gtk_combo_box_get_active_iter(GTK_COMBO_BOX(currencies_menu_from), &iter)) {
		gtk_tree_model_get(GTK_TREE_MODEL(currencies_store), &iter, 2, &uFrom, -1);
	}	
	if(gtk_combo_box_get_active_iter(GTK_COMBO_BOX(currencies_menu_to), &iter)) {
		gtk_tree_model_get(GTK_TREE_MODEL(currencies_store), &iter, 2, &uTo, -1);
	}
	
	if(uFrom && uTo) {
		//values
		const gchar *fromValue = gtk_entry_get_text(GTK_ENTRY(currencies_entry_from_val));
		const gchar *toValue = gtk_entry_get_text(GTK_ENTRY(currencies_entry_to_val));
		g_signal_handlers_block_matched((gpointer) currencies_entry_from_val, G_SIGNAL_MATCH_FUNC, 0, 0, NULL, (gpointer) on_currencies_entry_from_val_changed, NULL);
		g_signal_handlers_block_matched((gpointer) currencies_entry_to_val, G_SIGNAL_MATCH_FUNC, 0, 0, NULL, (gpointer) on_currencies_entry_to_val_changed, NULL);
		if(toFrom > 0) {
			MathStructure v_mstruct = CALCULATOR->convert(CALCULATOR->unlocalizeExpression(toValue, evalops.parse_options), uTo, uFrom, evalops);
			gtk_entry_set_text(GTK_ENTRY(currencies_entry_from_val), v_mstruct.print(printops).c_str());
		} else {
			MathStructure v_mstruct = CALCULATOR->convert(CALCULATOR->unlocalizeExpression(fromValue, evalops.parse_options), uFrom, uTo, evalops);
			gtk_entry_set_text(GTK_ENTRY(currencies_entry_to_val), v_mstruct.print(printops).c_str());
		}
		g_signal_handlers_unblock_matched((gpointer) currencies_entry_from_val, G_SIGNAL_MATCH_FUNC, 0, 0, NULL, (gpointer) on_currencies_entry_from_val_changed, NULL);
		g_signal_handlers_unblock_matched((gpointer) currencies_entry_to_val, G_SIGNAL_MATCH_FUNC, 0, 0, NULL, (gpointer) on_currencies_entry_to_val_changed, NULL);
	}
}


#ifdef __cplusplus
}
#endif


GtkWidget *create_currencies_dialog(void) {

	GtkWidget *dialog_vbox5;
	GtkWidget *currencies_table;
	GtkWidget *dialog_action_area5;
	GtkWidget *currencies_button_fetch_exchange_rates;
	GtkWidget *alignment1;
	GtkWidget *hbox1;
	GtkWidget *image1;
	GtkWidget *label1;
	GtkWidget *currencies_button_close;
	GtkTooltips *tooltips;

	tooltips = gtk_tooltips_new ();

	currencies_dialog = gtk_dialog_new ();
	gtk_container_set_border_width (GTK_CONTAINER (currencies_dialog), 5);
	gtk_window_set_title (GTK_WINDOW (currencies_dialog), _("Currency Converter"));
	gtk_window_set_resizable (GTK_WINDOW (currencies_dialog), FALSE);
	gtk_window_set_type_hint (GTK_WINDOW (currencies_dialog), GDK_WINDOW_TYPE_HINT_DIALOG);
	gtk_dialog_set_has_separator (GTK_DIALOG (currencies_dialog), FALSE);

	dialog_vbox5 = GTK_DIALOG (currencies_dialog)->vbox;
	gtk_widget_show (dialog_vbox5);

	currencies_table = gtk_table_new (2, 2, TRUE);
	gtk_widget_show (currencies_table);
	gtk_box_pack_start (GTK_BOX (dialog_vbox5), currencies_table, TRUE, TRUE, 0);
	gtk_container_set_border_width (GTK_CONTAINER (currencies_table), 12);
	gtk_table_set_row_spacings (GTK_TABLE (currencies_table), 6);
	gtk_table_set_col_spacings (GTK_TABLE (currencies_table), 6);

	currencies_entry_to_val = gtk_entry_new ();
	gtk_entry_set_alignment(GTK_ENTRY(currencies_entry_to_val), 1.0);
	gtk_widget_show (currencies_entry_to_val);
	gtk_table_attach (GTK_TABLE (currencies_table), currencies_entry_to_val, 1, 2, 1, 2, (GtkAttachOptions) (GTK_EXPAND | GTK_FILL), (GtkAttachOptions) (0), 0, 0);
	gtk_entry_set_text (GTK_ENTRY (currencies_entry_to_val), _("1.00"));

	currencies_entry_from_val = gtk_entry_new ();
	gtk_entry_set_alignment(GTK_ENTRY(currencies_entry_from_val), 1.0);
	gtk_widget_show (currencies_entry_from_val);
	gtk_table_attach (GTK_TABLE (currencies_table), currencies_entry_from_val, 1, 2, 0, 1, (GtkAttachOptions) (GTK_EXPAND | GTK_FILL), (GtkAttachOptions) (0), 0, 0);
	gtk_entry_set_text (GTK_ENTRY (currencies_entry_from_val), _("1.00"));

	dialog_action_area5 = GTK_DIALOG (currencies_dialog)->action_area;
	gtk_widget_show (dialog_action_area5);
	gtk_button_box_set_layout (GTK_BUTTON_BOX (dialog_action_area5), GTK_BUTTONBOX_END);

	currencies_button_fetch_exchange_rates = gtk_button_new ();
	gtk_widget_show (currencies_button_fetch_exchange_rates);
	gtk_dialog_add_action_widget (GTK_DIALOG (currencies_dialog), currencies_button_fetch_exchange_rates, GTK_RESPONSE_CLOSE);
	GTK_WIDGET_SET_FLAGS (currencies_button_fetch_exchange_rates, GTK_CAN_DEFAULT);
	gtk_button_set_focus_on_click (GTK_BUTTON (currencies_button_fetch_exchange_rates), FALSE);
	gtk_widget_set_sensitive(currencies_button_fetch_exchange_rates, canfetch);
	gtk_tooltips_set_tip (tooltips, currencies_button_fetch_exchange_rates, _("Fetch current exchange rates from the Internet (ECB)"), NULL);

	alignment1 = gtk_alignment_new (0.5, 0.5, 0, 0);
	gtk_widget_show (alignment1);
	gtk_container_add (GTK_CONTAINER (currencies_button_fetch_exchange_rates), alignment1);

	hbox1 = gtk_hbox_new (FALSE, 2);
	gtk_widget_show (hbox1);
	gtk_container_add (GTK_CONTAINER (alignment1), hbox1);

	image1 = gtk_image_new_from_stock ("gtk-refresh", GTK_ICON_SIZE_BUTTON);
	gtk_widget_show (image1);
	gtk_box_pack_start (GTK_BOX (hbox1), image1, FALSE, FALSE, 0);

	label1 = gtk_label_new_with_mnemonic (_("Update Rates"));
	gtk_widget_show (label1);
	gtk_box_pack_start (GTK_BOX (hbox1), label1, FALSE, FALSE, 0);

	currencies_button_close = gtk_button_new_from_stock ("gtk-close");
	gtk_widget_show (currencies_button_close);
	gtk_dialog_add_action_widget (GTK_DIALOG (currencies_dialog), currencies_button_close, GTK_RESPONSE_CLOSE); 
	GTK_WIDGET_SET_FLAGS (currencies_button_close, GTK_CAN_DEFAULT);
	gtk_tooltips_set_tip (tooltips, currencies_button_close, _("Close this window"), NULL);
	
	icon_pixbuf = gdk_pixbuf_new_from_xpm_data((const char**) icon_xpm);

	gtk_window_set_icon(GTK_WINDOW(currencies_dialog), icon_pixbuf);
	
	currencies_store = gtk_list_store_new(3, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_POINTER);
	gtk_tree_sortable_set_sort_func(GTK_TREE_SORTABLE(currencies_store), 0, string_sort_func, GINT_TO_POINTER(0), NULL);
	gtk_tree_sortable_set_sort_column_id(GTK_TREE_SORTABLE(currencies_store), 0, GTK_SORT_ASCENDING);
	
	currencies_menu_to = gtk_combo_box_new_with_model(GTK_TREE_MODEL(currencies_store));
	currencies_menu_from = gtk_combo_box_new_with_model(GTK_TREE_MODEL(currencies_store));
	GtkCellRenderer *cell = gtk_cell_renderer_text_new();
	gtk_cell_layout_pack_start(GTK_CELL_LAYOUT(currencies_menu_to), cell, TRUE);
	gtk_cell_layout_add_attribute(GTK_CELL_LAYOUT(currencies_menu_to), cell, "text", 0);
	cell = gtk_cell_renderer_text_new();
	gtk_cell_layout_pack_start(GTK_CELL_LAYOUT(currencies_menu_to), cell, FALSE);
	gtk_cell_layout_add_attribute(GTK_CELL_LAYOUT(currencies_menu_to), cell, "text", 1);
	cell = gtk_cell_renderer_text_new();
	gtk_cell_layout_pack_start(GTK_CELL_LAYOUT(currencies_menu_from), cell, TRUE);
	gtk_cell_layout_add_attribute(GTK_CELL_LAYOUT(currencies_menu_from), cell, "text", 0);
	cell = gtk_cell_renderer_text_new();
	gtk_cell_layout_pack_start(GTK_CELL_LAYOUT(currencies_menu_from), cell, FALSE);
	gtk_cell_layout_add_attribute(GTK_CELL_LAYOUT(currencies_menu_from), cell, "text", 1);
	gtk_widget_show(currencies_menu_to);
	gtk_widget_show(currencies_menu_from);
	gtk_table_attach(GTK_TABLE(currencies_table), currencies_menu_to, 0, 1, 1, 2, (GtkAttachOptions) (GTK_FILL), (GtkAttachOptions) (GTK_FILL), 0, 0);
	gtk_table_attach(GTK_TABLE(currencies_table), currencies_menu_from, 0, 1, 0, 1, (GtkAttachOptions) (GTK_FILL), (GtkAttachOptions) (GTK_FILL), 0, 0);
	
	gtk_widget_grab_focus(currencies_entry_from_val);
	
	g_signal_connect ((gpointer) currencies_menu_from, "changed", G_CALLBACK (on_currencies_menu_from_changed), NULL);
	g_signal_connect ((gpointer) currencies_menu_to, "changed", G_CALLBACK (on_currencies_menu_to_changed), NULL);
	g_signal_connect_swapped ((gpointer) currencies_dialog, "delete_event", G_CALLBACK (close_app), NULL);
	g_signal_connect_swapped ((gpointer) currencies_dialog, "destroy_event", G_CALLBACK (close_app), NULL);
	g_signal_connect ((gpointer) currencies_entry_to_val, "changed", G_CALLBACK (on_currencies_entry_to_val_changed), NULL);
	g_signal_connect ((gpointer) currencies_entry_from_val, "changed", G_CALLBACK (on_currencies_entry_from_val_changed), NULL);
	g_signal_connect ((gpointer) currencies_button_fetch_exchange_rates, "clicked", G_CALLBACK (on_currencies_button_fetch_exchange_rates_clicked), NULL);
	g_signal_connect ((gpointer) currencies_button_close, "clicked", G_CALLBACK (on_currencies_button_close_clicked), NULL);

	return currencies_dialog;
	
}

void load_preferences() {

	evalops.approximation = APPROXIMATION_APPROXIMATE;
	evalops.parse_options.angle_unit = ANGLE_UNIT_RADIANS;
	evalops.structuring = STRUCTURING_NONE;
	evalops.parse_options.dot_as_separator = CALCULATOR->default_dot_as_separator;
	printops.number_fraction_format = FRACTION_DECIMAL;
	printops.lower_case_e = true;
	printops.min_decimals = 2;
	printops.max_decimals = 2;
	printops.use_min_decimals = true;
	printops.use_max_decimals = true;
	
	fetch_exchange_rates_at_startup = false;
	wget_args = "--quiet --tries=1";
	initial_from_currency = "";
	initial_to_currency = "";
	first_time = false;
	gchar *gstr = g_build_filename(g_get_home_dir(), ".qalculate", NULL);
	DIR *dir = opendir(gstr);
	g_free(gstr);
	if(!dir) {
		first_qalculate_run = true;
		first_time = true;
		return;
	} else {
		first_qalculate_run = false;
		closedir(dir);
	}
	int version_numbers[] = {0, 9, 3};
	FILE *file = NULL;
	gchar *gstr2 = g_build_filename(g_get_home_dir(), ".qalculate", "qalculate-currency.cfg", NULL);
	file = fopen(gstr2, "r");
	g_free(gstr2);
	if(file) {
		char line[10000];
		string stmp, svalue, svar;
		size_t i;
		int v;
		while(true) {
			if(fgets(line, 10000, file) == NULL)
				break;
			stmp = line;
			remove_blank_ends(stmp);
			if((i = stmp.find_first_of("=")) != string::npos) {
				svar = stmp.substr(0, i);
				remove_blank_ends(svar);
				svalue = stmp.substr(i + 1, stmp.length() - (i + 1));
				remove_blank_ends(svalue);
				v = s2i(svalue);
				if(svar == "version") {
					parse_qalculate_version(svalue, version_numbers);
				} else if(svar == "fetch_exchange_rates_at_startup") {
					fetch_exchange_rates_at_startup = v;
				} else if(svar == "wget_args") {
					wget_args = svalue;
				} else if(svar == "dot_as_separator") {
					evalops.parse_options.dot_as_separator = v;
				} else if(svar == "from_currency") {
					initial_from_currency = svalue;
				} else if(svar == "to_currency") {
					initial_to_currency = svalue;
				}
			}
		}
	} else {
		first_time = true;
	}
}


void save_preferences() {
	FILE *file = NULL;
	gchar *gstr = g_build_filename(g_get_home_dir(), ".qalculate", NULL);
	mkdir(gstr, S_IRWXU);
	g_free(gstr);
	gchar *gstr2 = g_build_filename(g_get_home_dir(), ".qalculate", "qalculate-currency.cfg", NULL);
	file = fopen(gstr2, "w+");
	g_free(gstr2);
	fprintf(file, "\n[General]\n");
	fprintf(file, "version=%s\n", VERSION);	
	fprintf(file, "fetch_exchange_rates_at_startup=%i\n", fetch_exchange_rates_at_startup);
	fprintf(file, "wget_args=%s\n", wget_args.c_str());
	fprintf(file, "dot_as_separator=%i\n", evalops.parse_options.dot_as_separator);
	Unit *uFrom = NULL;
	Unit *uTo = NULL;
	GtkTreeIter iter;
	if(gtk_combo_box_get_active_iter(GTK_COMBO_BOX(currencies_menu_from), &iter)) {
		gtk_tree_model_get(GTK_TREE_MODEL(currencies_store), &iter, 2, &uFrom, -1);
	}	
	if(gtk_combo_box_get_active_iter(GTK_COMBO_BOX(currencies_menu_to), &iter)) {
		gtk_tree_model_get(GTK_TREE_MODEL(currencies_store), &iter, 2, &uTo, -1);
	}
	if(uFrom) fprintf(file, "from_currency=%s\n", uFrom->referenceName().c_str());
	if(uTo) fprintf(file, "to_currency=%s\n", uTo->referenceName().c_str());
	fclose(file);
}


int main (int argc, char **argv) {

#ifdef ENABLE_NLS
	bindtextdomain (GETTEXT_PACKAGE, PACKAGE_LOCALE_DIR);
	bind_textdomain_codeset (GETTEXT_PACKAGE, "UTF-8");
	textdomain (GETTEXT_PACKAGE);
#endif

	gtk_init(&argc, &argv);

	struct lconv *lc = localeconv();
	bool decimal_comma = (strcmp(lc->mon_decimal_point, ",") == 0);
	
	//create the almighty Calculator object
	new Calculator();

	if(decimal_comma) CALCULATOR->useDecimalComma();
	else CALCULATOR->useDecimalPoint();

	CALCULATOR->default_dot_as_separator = strcmp(lc->mon_thousands_sep, ".") == 0;

	load_preferences();
	
	canfetch = CALCULATOR->canFetch();
	
	create_currencies_dialog();
	while(gtk_events_pending()) gtk_main_iteration();
	
	//exchange rates
	FILE *file = NULL;
	file = fopen(CALCULATOR->getExchangeRatesFileName().c_str(), "r");
	if(!file && canfetch) {
		GtkWidget *edialog = gtk_message_dialog_new(GTK_WINDOW(currencies_dialog), GTK_DIALOG_DESTROY_WITH_PARENT, GTK_MESSAGE_WARNING, GTK_BUTTONS_YES_NO, _("You need to download exchange rates to be able to convert between different currencies. You can later get current exchange rates by clicking the \"Update Rates\" button.\n\nDo you want to fetch exchange rates now from the Internet?"));
		int question_answer = gtk_dialog_run(GTK_DIALOG(edialog));
		gtk_widget_destroy(edialog);
		while(gtk_events_pending()) gtk_main_iteration();
		if(question_answer == GTK_RESPONSE_YES) {
			fetch_exchange_rates(5);
			while(gtk_events_pending()) gtk_main_iteration();
		}
	} else if(fetch_exchange_rates_at_startup && canfetch) {
		fetch_exchange_rates(5);
		while(gtk_events_pending()) gtk_main_iteration();
	}
	if(file) fclose(file);

	CALCULATOR->loadExchangeRates();

	//load global definitions
	if(!CALCULATOR->loadGlobalCurrencies()) {
		g_print(_("Failed to load global definitions!\n"));
	}

	update_currency_menus();
	
	convert_currencies();
	
	gtk_widget_show(currencies_dialog);
	
	gtk_main();
	
	return 0;
}
